/**
 * AlternateMode
 * Handles sensors in alternate mode.
 *
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var AlternateMode = function()
{
    var sensors   = this.sensors;
    var self      = this;
    var transport = app.transportButtons.hostTransport;

    sensors.solo         = new AlwaysOnSensor( 'solo' );
    sensors.soloEx       = new AlwaysOnSensor( 'soloEx' );
    sensors.mute         = new AlwaysOnSensor( 'mute' );
    sensors.activate     = new AlwaysOnSensor( 'activate' );
    sensors.deviceDown   = new Sensor( 'deviceDown' );
    sensors.deviceUp     = new Sensor( 'deviceUp' );
    sensors.scalesEnable = new AlwaysOnSensor( 'scalesEnable' );
    sensors.splitMode    = new AlwaysOnSensor( 'splitMode' );

    sensors.trackPan     = new Rotary( 0, 127, 'trackPan' );
    sensors.tempo        = new Rotary( 0, 127, 'tempo' );

    sensors.arm          = new Sensor( 'arm', 0 );
    sensors.click        = new Sensor( 'click', 0 );
    sensors.loop         = new Sensor( 'loop', 0 );

    this.addListener( 'solo', 'sensorOn', this, function() {
        app.cursorTrack.getSolo().toggle( false );
        app.modeButton.onAlternateEvent();
    });
    this.addListener( 'soloEx', 'sensorOn', this, function() {
        app.cursorTrack.getSolo().toggle( true );
        app.modeButton.onAlternateEvent();
    });
    this.addListener( 'mute', 'sensorOn', this, function() {
        app.cursorTrack.getMute().toggle();
        app.modeButton.onAlternateEvent();
    });
    this.addListener( 'activate', 'sensorOn', this, function() {
        app.cursorTrack.isActivated().toggle();
        app.modeButton.onAlternateEvent();
    });

    // device selectors
    this.addListener( 'deviceDown', 'sensorOn', this, function() {
        app.device.selectPrevious();
        app.modeButton.onAlternateEvent();
    });
    this.addListener( 'deviceUp', 'sensorOn', this, function() {
        app.device.selectNext();
        app.modeButton.onAlternateEvent();
    });
    app.device.addCanSelectPreviousObserver( function( canSelect ) {
        self.sensors.deviceDown.setValue( canSelect, self.isActive() );
    });
    app.device.addCanSelectNextObserver( function( canSelect ) {
        self.sensors.deviceUp.setValue( canSelect, self.isActive() );
    });

    // arm
    this.addListener( 'arm', 'sensorOn', this, function() {
        app.cursorTrack.getArm().toggle();
        app.modeButton.onAlternateEvent();
    });
    app.cursorTrack.getArm().addValueObserver( function( armed ) {
        sensors.arm.setValue( armed, self.isActive() );
    });

    // click
    this.addListener( 'click', 'sensorOn', this, function( sensor ) {
        transport.setClick( !sensor.getValue() );
        app.modeButton.onAlternateEvent();
    });
    transport.addClickObserver( function( value ) {
        sensors.click.setValue( value, self.isActive() );
    });

    // loop
    this.addListener( 'loop', 'sensorOn', this, function( sensor ) {
        transport.setLoop( !sensor.getValue() );
        app.modeButton.onAlternateEvent();
    });
    transport.addIsLoopActiveObserver( function( value ) {
        sensors.loop.setValue( value, self.isActive() );
    });

    // track pan
    var trackPan = app.cursorTrack.getPan();
    trackPan.addValueObserver( 128, function( pan ) {
        sensors.trackPan.setValue( 128 - pan, self.isActive() );
    });
    this.addListener( 'trackPan', 'sensorMove', this, function( sensor ) {
        trackPan.set( 128 - sensor.getValue(), 128 );
        app.modeButton.onAlternateEvent();
    });

    // tempo
    this.addListener( 'tempo', 'sensorMove', this, function( sensor ) {
        transport.getTempo().setRaw( sensor.getValue()*2 );
        app.modeButton.onAlternateEvent();
    });
    transport.getTempo().addValueObserver( Math.floor( 128*2 ), function( value ) {
        sensors.tempo.setValue( Math.min( 127, value ), self.isActive() );  // Bitwig tempo goes up to 666 (=128*5.2). Here the range is limited to 248.
    });

    // scales enable
    this.addListener( 'scalesEnable', 'sensorOn', this, function() {
        var scalesEnabled = !app.preferences.scalesEnabled;
        app.preferences.scalesEnabledSetting.set( scalesEnabled ? 'On' : 'Off' );
        app.modeButton.onAlternateEvent();
    });
    // split mode
    this.addListener( 'splitMode', 'sensorOn', this, function() {
        var mode = !app.preferences.splitMode;
        app.preferences.splitModeSetting.set( mode ? 'On' : 'Off' );
        app.modeButton.onAlternateEvent();
    });

};
AlternateMode.prototype = new SensorGroup();
AlternateMode.prototype.constructor = AlternateMode;


AlternateMode.prototype.onNoteOn = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.solo.note         : this.sensors.solo.onPress(); break;
        case app.layout.soloEx.note       : this.sensors.soloEx.onPress(); break;
        case app.layout.mute.note         : this.sensors.mute.onPress(); break;
        case app.layout.activate.note     : this.sensors.activate.onPress(); break;
        case app.layout.deviceDown.note   : this.sensors.deviceDown.onPress(); break;
        case app.layout.deviceUp.note     : this.sensors.deviceUp.onPress(); break;
        case app.layout.scalesEnable.note : this.sensors.scalesEnable.onPress(); break;
        case app.layout.splitMode.note    : this.sensors.splitMode.onPress(); break;
        case app.layout.arm.note          : this.sensors.arm.onPress(); break;
        case app.layout.click.note        : this.sensors.click.onPress(); break;
        case app.layout.loop.note         : this.sensors.loop.onPress(); break;
    }
};


AlternateMode.prototype.onNoteOff = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.solo.note         : this.sensors.solo.onRelease(); break;
        case app.layout.soloEx.note       : this.sensors.soloEx.onRelease(); break;
        case app.layout.mute.note         : this.sensors.mute.onRelease(); break;
        case app.layout.activate.note     : this.sensors.activate.onRelease(); break;
        case app.layout.deviceDown.note   : this.sensors.deviceDown.onRelease(); break;
        case app.layout.deviceUp.note     : this.sensors.deviceUp.onRelease(); break;
        case app.layout.scalesEnable.note : this.sensors.scalesEnable.onRelease(); break;
        case app.layout.splitMode.note    : this.sensors.splitMode.onRelease(); break;
        case app.layout.arm.note          : this.sensors.arm.onRelease(); break;
        case app.layout.click.note        : this.sensors.click.onRelease(); break;
        case app.layout.loop.note         : this.sensors.loop.onRelease(); break;
        case app.layout.tempo.note        : this.sensors.tempo.onRelease(); break;
        case app.layout.trackPan.note     : this.sensors.trackPan.onRelease(); break;
    }
};


AlternateMode.prototype.onController = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.trackPan.cc : this.sensors.trackPan.onMove( midi.data2 ); break;
        case app.layout.tempo.cc    : this.sensors.tempo.onMove( midi.data2 ); break;
    }
}


AlternateMode.prototype.permittedModes = ['alternate'];

