

/**
 * App
 * Application class.
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */


load( 'e3_utils.js' );
load( 'e3_midi_data.js' );
load( 'e3_preferences.js' );
load( 'e3_messages.js' );
load( 'e3_midi_monitor.js' );
load( 'e3_sensors.js' );
load( 'e3_pads.js' );
load( 'e3_transport.js' );
load( 'e3_notes.js' );
load( 'e3_clip_buttons.js' );
load( 'e3_track_button.js' );
load( 'e3_pitchbend.js' );
load( 'e3_volume.js' );
load( 'e3_sliders.js' );
load( 'e3_alternate_mode.js' );
load( 'e3_router.js' );
load( 'e3_mode_button.js' );


var App = function() {};


App.prototype.init = function()
{
    this.preferences = new Preferences();
    this.messages    = new Messages();
    this.message( 'Initializing' );

    var notifications = host.getNotificationSettings();
    notifications.setShouldShowChannelSelectionNotifications( true );
    notifications.setShouldShowDeviceLayerSelectionNotifications( true );
    notifications.setShouldShowDeviceSelectionNotifications( true );
    notifications.setShouldShowMappingNotifications( true );
    notifications.setShouldShowPresetNotifications( false );
    notifications.setShouldShowSelectionNotifications( true );
    notifications.setShouldShowTrackSelectionNotifications( true );
    notifications.setShouldShowValueNotifications( false );

    this.inPort  = host.getMidiInPort(0);
    this.outPort = host.getMidiOutPort(0);

    this.noteInput = this.inPort.createNoteInput( 'Quneo-E3', '8?????', '9?????' );
    this.noteInput.setShouldConsumeEvents( false );

    this.cursorTrack      = host.createArrangerCursorTrack( 8, 0 );
    this.masterTrack      = host.createMasterTrack( 8 );
    this.clipLauncher     = host.createTrackBank( 4, 2, 4 );
    this.device           = host.createEditorCursorDevice();

    // Not working since 1.3.6
    //this.cursorTrack      = host.createArrangerCursorTrack( 8, 0 );
    //this.device           = this.cursorTrack.createCursorDevice();

    this.modeButton       = new ModeButton();
    this.transportButtons = new TransportButtons();
    this.monitor          = new MidiMonitor();
    this.clipButtons      = new ClipButtonGroup();
    this.trackButtons     = new TrackButton();
    this.volume           = new VolumeRotaryGroup();
    this.pitchbend        = new PitchbendSlider();
    this.notes            = new Notes();
    this.pads             = new Pads();
    this.router           = new Router();
    this.alternateMode    = new AlternateMode();

    this.updateLeds();
    this.notes.updateTranslationTable();
    this.inPort.setMidiCallback( utils.bind( this.onMidi, this ));
    this.inPort.setSysexCallback( utils.bind( this.onSysEx, this ));
};


App.prototype.onMidi = function( status, data1, data2 )
{
    var midi = this.prepareMidi( status, data1, data2 );

    this.monitor.onMidi( midi );
    this.trackButtons.onMidi( midi );
    this.notes.onMidi( midi );
    this.clipButtons.onMidi( midi );
    this.pads.onMidi( midi );
    this.router.onMidi( midi );
    this.transportButtons.onMidi( midi );
    this.volume.onMidi( midi );
    this.pitchbend.onMidi( midi );
    this.alternateMode.onMidi( midi );
    this.modeButton.onMidi( midi );
};


App.prototype.onSysEx = function( sysex ) {
    if( sysex.matchesHexPattern( 'f07e00060300015f1e0000001e1200??f7' )) {    // Quneo load pattern
        this.updateLeds();
    }
};


App.prototype.updateLeds = function()
{
    this.modeButton.updateLeds();
    this.transportButtons.updateLeds();
    this.trackButtons.updateLeds();
    this.pitchbend.updateLeds();
    this.clipButtons.updateLeds();
    this.volume.updateLeds();
    this.notes.updateLeds();
    this.router.updateLeds();
    this.pads.updateLeds();
};


// TODO: put this function to e3_transposer.js
//
App.prototype.prepareMidi = function( status, data1, data2 )
{
    var midi      = new MidiData( status, data1, data2 );
    midi.mode     = this.getMode();
    midi.mainMode = this.getMainMode();
    midi.root     = this.getPadRoot();
    midi.padIndex = midi.data1 - midi.root;

    if( midi.data1 >= midi.root && midi.data1 < midi.root + 16) {
        switch( midi.mainMode ) {
            case 'notes':
            case 'clips': midi.padMode = midi.mainMode; break;
            case 'split':
                if( midi.data1 < midi.root + 8 )       midi.padMode = 'notes';
                else if( midi.data1 >= midi.root + 8 ) midi.padMode = 'clips';
                break;
        }
    }
    return midi;
};


App.prototype.getPadRoot      = function() { return this.layout.rootNote.note; };
App.prototype.getMode         = function() { return this.modeButton.getMode(); };
App.prototype.getMainMode     = function() { return this.modeButton.getMainMode(); };
App.prototype.isAlternateMode = function() { return this.modeButton.isAlternateMode(); }
App.prototype.getScales       = function() { return Notes.scales; };

App.prototype.exit = function() {
    this.message( 'Exiting' );
};

App.prototype.message = function( msg ) {
    this.messages.print( msg );
};

/**
 * Sends the given midi event to Quneo.
 * @param status
 * @param data1
 * @param data2
 */
App.prototype.midiToQuneo = function( status, data1, data2, debugInfo )
{
    if( data1 < 0 || data1 > 127 || data2 < 0 || data2 > 127) {
        throw 'App.midiToQuneo value out of range ('+data1+', '+data2+'), caller='+debugInfo;
    }
    this.outPort.sendMidi( status, data1, data2 );
};

/**
 * Sends the given sysex data to Quneo
 * @param sysex
 */
App.prototype.sysexToQuneo = function( sysex ) {
    this.outPort.sendSysex( sysex );
};


/**
 * Converts the given value to a pitch bend event and sends it to Bitwig.
 * @param value
 */
//App.prototype.pitchbendToBitwig = function( value )
//{
//    var data = MidiData.encode14BitValue( value );
//    this.noteInput.sendRawMidiEvent( 0xE0, data.msb, data.lsb );
//};


/**
 * Sends a midi event to Bitwig
 * @param status
 * @param data1
 * @param data2
 */
App.prototype.midiToBitwig = function( status, data1, data2 )
{
    this.noteInput.sendRawMidiEvent( status, data1, data2 );
};


App.prototype.notify = function( msg )
{
    this.message( msg );
    host.showPopupNotification( msg );
};


App.ledCodes = {
    pad1         : { note: 36, ledNum: 0,  ledStatus: 144 },
    playButton   : { note: 31, ledNum: 35, ledStatus: 144 },
    stopButton   : { note: 30, ledNum: 34, ledStatus: 144 },
    recordButton : { note: 29, ledNum: 33, ledStatus: 144 },

    upButton1    : { note: 20, ledNum: 46, ledStatus: 144 },
    downButton1  : { note: 21, ledNum: 47, ledStatus: 144 },
    upButton2    : { note: 22, ledNum: 48, ledStatus: 144 },
    downButton2  : { note: 23, ledNum: 49, ledStatus: 144 },

    upButton1    : { note: 20, ledNum: 46, ledStatus: 144 },
    downButton1  : { note: 21, ledNum: 47, ledStatus: 144 },
    upButton2    : { note: 22, ledNum: 48, ledStatus: 144 },
    downButton2  : { note: 23, ledNum: 49, ledStatus: 144 },

    leftButton1  : { note: 12, ledNum: 36, ledStatus: 144 },
    rightButton1 : { note: 13, ledNum: 37, ledStatus: 144 },
    leftButton2  : { note: 14, ledNum: 38, ledStatus: 144 },
    rightButton2 : { note: 15, ledNum: 39, ledStatus: 144 },
    leftButton3  : { note: 16, ledNum: 40, ledStatus: 144 },
    rightButton3 : { note: 17, ledNum: 41, ledStatus: 144 },
    leftButton4  : { note: 18, ledNum: 42, ledStatus: 144 },
    rightButton4 : { note: 19, ledNum: 43, ledStatus: 144 },

    rhombusButton: { note: 32,         ledNum: 44, ledStatus: 144 },
    rotary1      : { note: 9,  cc: 9,  ledNum: 6,  ledStatus: 176 },
    rotary2      : { note: 10, cc: 10, ledNum: 7,  ledStatus: 176 },
    longSlider   : { note: 11, cc: 11, ledNum: 5,  ledStatus: 176 },

    vslider1     : { note: 1, cc: 1,  ledNum: 1,  ledStatus: 176 },
    vslider2     : { note: 2, cc: 2,  ledNum: 2,  ledStatus: 176 },
    vslider3     : { note: 3, cc: 3,  ledNum: 3,  ledStatus: 176 },
    vslider4     : { note: 4, cc: 4,  ledNum: 4,  ledStatus: 176 },
    hslider1     : { note: 5, cc: 5,  ledNum: 11, ledStatus: 176 },
    hslider2     : { note: 6, cc: 6,  ledNum: 10, ledStatus: 176 },
    hslider3     : { note: 7, cc: 7,  ledNum: 9,  ledStatus: 176 },
    hslider4     : { note: 8, cc: 8,  ledNum: 8,  ledStatus: 176 }
};


App.prototype.layout = {
    rootNote        : App.ledCodes.pad1,
    transportStart  : App.ledCodes.playButton,
    transportStop   : App.ledCodes.stopButton,
    transportRecord : App.ledCodes.recordButton,

    octaveUp        : App.ledCodes.upButton1,
    octaveDown      : App.ledCodes.downButton1,
    semitoneUp      : App.ledCodes.upButton2,
    semitoneDown    : App.ledCodes.downButton2,

    clipsUp         : App.ledCodes.upButton1,
    clipsDown       : App.ledCodes.downButton1,
    clipsLeft       : App.ledCodes.upButton2,
    clipsRight      : App.ledCodes.downButton2,

    trackDown       : App.ledCodes.leftButton1,
    trackUp         : App.ledCodes.rightButton1,
    topTargetDown   : App.ledCodes.leftButton2,
    topTargetUp     : App.ledCodes.rightButton2,
    subTargetDown   : App.ledCodes.leftButton3,
    subTargetUp     : App.ledCodes.rightButton3,
    scaleDown       : App.ledCodes.leftButton4,
    scaleUp         : App.ledCodes.rightButton4,

    toggleMode      : App.ledCodes.rhombusButton,
    trackVolume     : App.ledCodes.rotary1,
    masterVolume    : App.ledCodes.rotary2,
    pitchbend       : App.ledCodes.longSlider,

    slider1         : App.ledCodes.vslider1,
    slider2         : App.ledCodes.vslider2,
    slider3         : App.ledCodes.vslider3,
    slider4         : App.ledCodes.vslider4,
    slider5         : App.ledCodes.hslider1,
    slider6         : App.ledCodes.hslider2,
    slider7         : App.ledCodes.hslider3,
    slider8         : App.ledCodes.hslider4,

    // Alternate
    solo            : App.ledCodes.leftButton1,
    soloEx          : App.ledCodes.rightButton1,
    mute            : App.ledCodes.leftButton2,
    activate        : App.ledCodes.rightButton2,
    deviceDown      : App.ledCodes.leftButton3,
    deviceUp        : App.ledCodes.rightButton3,
    scalesEnable    : App.ledCodes.leftButton4,
    splitMode       : App.ledCodes.rightButton4,

    arm             : App.ledCodes.recordButton,
    click           : App.ledCodes.stopButton,
    loop            : App.ledCodes.playButton,

    trackPan        : App.ledCodes.rotary1,
    tempo           : App.ledCodes.rotary2,
};


//App.prototype.scales = [
//    { name: 'Chromatic',          notes: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11] },
//    { name: 'Major',              notes: [0, 2, 4, 5, 7, 9, 11] },
//    { name: 'Melodic Minor',      notes: [0, 2, 3, 5, 7, 9, 11] },
//    { name: 'Harmonic Minor',     notes: [0, 2, 3, 5, 7, 8, 11] },
//    { name: 'Gypsy Minor',        notes: [0, 2, 3, 6, 7, 8, 11] },
//    { name: 'Major Pentatonic',   notes: [0, 2, 4, 7, 9] },
//    { name: 'Minor Pentatonic',   notes: [0, 3, 5, 7, 10] },
//    { name: 'Minor-6 Pentatonic', notes: [0, 3, 5, 7, 9] },
//    { name: 'Blues',              notes: [0, 3, 5, 6, 7, 10] },
//    { name: 'Japanese',           notes: [0, 1, 5, 7, 8] },
//    { name: 'Hindu',              notes: [0, 2, 5, 7, 8, 11] },
//    { name: 'Diminished',         notes: [0, 1, 3, 4, 6, 7, 9, 10] },
//    { name: 'Whole Tone',         notes: [0, 2, 4, 6, 8, 10] },
//];




