/**
 * ModeButton
 *
 * Binds the sensors to different functionality for each mode. Modes are:
 *
 * - notes: pads play notes, the up/down buttons shift octave and semitone
 * - clips: pads launch clips, the up/down buttons scroll tracks and scenes
 * - split: the lower half of the pads is in note mode, the upper half in clip mode.
 * - alternate: the left/right buttons have different functionality, up/down buttons as in clip mode
 *
 * The rhombus button switches the modes. While it is pressed the mode is alternate mode.
 * After release it switches to next mode in circle motion.
 * If the alternate mode lasted more than two seconds or one of the alternate buttons was pressed during
 * alternate mode, the previous mode will be restored after release.
 *
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var ModeButton = function()
{
    this.canToggle = true;
    this.buttonPressed = false;
    var ledColorMap = {
        notes:     [ 127, 0 ],
        clips:     [ 0, 127 ],
        split:     [ 127, 50 ],
        alternate: [ 127, 50 ]
    };
    this.button = new RhombusButton( 'toggleMode', 'notes', ledColorMap );
    this.button.addListener( 'sensorOn',  this, this.setAlternateMode );
    this.button.addListener( 'sensorOff', this, this.toggleMode );

    app.preferences.addListener( 'mode', this, function() {
        this.setMode( app.preferences.mode.toLowerCase() );
    });
    app.preferences.addListener( 'splitMode', this, function() {
        this.setMode( app.preferences.mode.toLowerCase() );
    });
};
ModeButton.prototype = new SensorGroup();
ModeButton.prototype.constructor = ModeButton;


/**
 * Toggles between notes and clips mode.
 */
ModeButton.prototype.toggleMode = function()
{
    if( this.canToggle === true )
    {
        var mode = (app.preferences.mode === 'clips') ? 'notes' : 'clips';
        app.preferences.modeSetting.set( mode.capitalize() );
        app.notify( 'Mode: ' + mode.capitalize() );
    }
    else {
        this.button.setValue( this.getMainMode() );
        app.updateLeds();
    }
    this.canToggle = true;
    this.buttonPressed = false;
};


/**
 * Sets the given mode.
 * @param mode
 * @param notify If true the application will be notified about the mode.
 */
ModeButton.prototype.setMode = function( mode, notify )
{
    if( mode === 'clips' && app.preferences.splitMode ) mode = 'split';
    this.button.setValue( mode );

    if( notify ) app.notify( 'Mode: ' + mode.capitalize() );
    app.updateLeds();
};


/**
 * Sets alternate mode.
 */
ModeButton.prototype.setAlternateMode = function()
{
    this.button.setValue( 'alternate' );
    this.buttonPressed = true;
    this.canToggle = true;
    app.notify( 'Mode: Alternate' );

    app.alternateMode.updateLeds();
    app.clipButtons.updateLeds();
    app.notes.onEnterAlternateMode();

    host.scheduleTask( utils.bind( function() {
        if( this.getMode() === 'alternate' ) {
            this.canToggle = false;
        }
    }, this ), null, 1000 );
};


ModeButton.prototype.getMode = function() {
    return this.button.getValue();
};


ModeButton.prototype.getMainMode = function()
{
    var mode = app.preferences.mode;
    if( mode === 'clips' && app.preferences.splitMode ) mode = 'split';
    return mode;
};


ModeButton.prototype.isAlternateMode = function() {
    return this.buttonPressed;
};


ModeButton.prototype.updateLeds = function() {
    this.button.sendLed();
};


/**
 * Sets the canToggle-flag to false, so the mode will not be toggled when the button is released.
 */
ModeButton.prototype.onAlternateEvent = function() {
    this.canToggle = false;
};


ModeButton.prototype.onNoteOn = function( midi )
{
    if( midi.data1 === app.layout.toggleMode.note ) {
        this.button.onPress();
    }
};


ModeButton.prototype.onNoteOff = function( midi )
{
    if( midi.data1 === app.layout.toggleMode.note ) {
        this.button.onRelease();
    }
};
