/**
 * Notes
 * Handles all midi input. Includes pitch handling, scales and transposition.
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var Notes = function()
{
    SensorGroup.prototype.constructor.call( this );

    this.noteTable = initArray( -1, 128 );
    this.holdNotes = initArray( -1, 16 );  // tracks which notes the hold pads are playing in order to send correct note-off
    this.playNotes = initArray( -1, 16 );  // tracks any playing notes

    this.sensors.octave   = new RangedSensorPair( 0, -4, 6, 'octaveUp', 'octaveDown', true );
    this.sensors.semitone = new RangedSensorPair( 0, -12, 12, 'semitoneUp', 'semitoneDown', true );
    this.sensors.scale    = new CircleSensorPair( 0, Notes.scales.length, 'scaleUp', 'scaleDown' );

    this.addListener( 'octave', 'sensorOn', this, function( sensor ) {
        this.setOctave( sensor.value );
    });
    this.addListener( 'semitone', 'sensorOn', this, function( sensor ) {
        this.setSemitones( sensor.value );
    });
    this.addListener( 'scale', 'sensorOn', this, function( sensor ) {
        this.setScale( sensor.value );
    });

    app.preferences.addListener( 'octave',         this, this.onPreferences );
    app.preferences.addListener( 'semitone',       this, this.onPreferences );
    app.preferences.addListener( 'scale',          this, this.onPreferences );
    app.preferences.addListener( 'scalesEnabled',  this, this.update );
    app.preferences.addListener( 'mode',           this, this.update );

    Observer.makeObserver( this );
    app.noteInput.setKeyTranslationTable( initArray( -1, 128 ));  // handle midi notes self, not Bitwig API
};
Notes.prototype = new SensorGroup();
Notes.prototype.constructor = Notes;


Notes.prototype.update = function()
{
    this.updateTranslationTable();
    this.fire( 'notesUpdate' );
    this.updateLeds();
};


/**
 * Updates the key translation table
 */
Notes.prototype.updateTranslationTable = function()
{
    this.noteTable  = initArray( -1, 128 );
    var context = this.buildTranslationContext();

    if( context.mode === 'notes' || context.mode === 'split' ) {
        for( var i = context.padRoot; i < context.maxIndex; i++ ) {
            this.noteTable[i] = this.translate( i, context );
        }
    }
};


/**
 * Translates the given midi numbers according to the current scale, scaleEnabled, modalTranspose and transpose state.
 * @param midiNum
 * @param context
 * @returns {number}
 */
Notes.prototype.translate = function( midiNum, context )
{
    context = context || this.buildTranslationContext();

    if( app.preferences.scalesEnabled ) {
        var offset  = midiNum - context.padRoot;
        var degree  = Math.abs( offset % context.scaleLength );
        var octave  = Math.floor( offset / context.scaleLength );
        var note    = context.scale[ degree ];
        var result  = context.padRoot + note + 12 * octave + context.transpose;

        return Math.min( 127, Math.max( 0, result ));
    }
    else {
        return Math.min( 127, midiNum + context.transpose );
    }
};


/**
 * Builds the context for this.translate.
 */
Notes.prototype.buildTranslationContext = function()
{
    var mode = app.getMainMode();
    var padRoot = app.getPadRoot();
    return {
        mode        : mode,
        padRoot     : padRoot,
        scale       : this.getScale(),
        scaleLength : this.getScale().length,
        maxIndex    : mode === 'split' ? padRoot + 8 : 128,
        transpose   : this.getOctave() * 12 + this.getSemitones()
    };
};


Notes.prototype.onNoteOn = function( midi )
{
    if( app.getMode() !== 'alternate' ) {
        switch( midi.data1 ) {
            case app.layout.scaleUp.note   : this.sensors.scale.increase(); break;
            case app.layout.scaleDown.note : this.sensors.scale.decrease(); break;
        }
    }

    switch( midi.data1 ) {
        case app.layout.octaveUp.note     : this.sensors.octave.increase(); break;
        case app.layout.octaveDown.note   : this.sensors.octave.decrease(); break;
        case app.layout.semitoneUp.note   : this.sensors.semitone.increase(); break;
        case app.layout.semitoneDown.note : this.sensors.semitone.decrease(); break;
    }

    var note = this.noteTable[midi.data1];
    var maxIndex = midi.mainMode === 'split' ? 8 : 16;

    if( midi.padIndex >= 0 && midi.padIndex < maxIndex && note > -1 )
    {
        if( midi.mainMode === 'notes' || midi.mainMode === 'split' )
        {
            this.playNotes[midi.padIndex] = note;

            if( app.preferences.holdPads[midi.padIndex] > 0 )   // pad has hold mode enabled
            {
                var holdNote = this.holdNotes[midi.padIndex];
                if( holdNote >= 0 ) {
                    this.clearHoldNote( midi.padIndex );        // clear hold
                }
                else {
                    this.holdNotes[midi.padIndex] = note;       // set hold
                    app.midiToBitwig( 144, note, midi.data2 );
                }
            }
            else {
                app.midiToBitwig( 144, note, midi.data2 );
            }

            if( app.getMode() === 'alternate' )  // toggle hold mode for note pads
            {
                var holdPads = app.preferences.holdPads;
                var hold     = (holdPads[midi.padIndex] > 0) ? 0 : 1;
                if( hold === 0 ) {
                    this.clearHoldNote( midi.padIndex );
                }
                this.setHoldNote( midi.padIndex, hold );
            }
        }
    }
};


Notes.prototype.onNoteOff = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.octaveUp.note     :
        case app.layout.octaveDown.note   : this.sensors.octave.onRelease(); break;
        case app.layout.semitoneUp.note   :
        case app.layout.semitoneDown.note : this.sensors.semitone.onRelease(); break;
        case app.layout.scaleUp.note      :
        case app.layout.scaleDown.note    : this.sensors.scale.onRelease(); break;
    }

    if( midi.padIndex >= 0 && midi.padIndex < 16 && this.noteTable[midi.data1] > -1 )
    {
        this.padPressed = false;

        if( this.holdNotes[midi.padIndex] === -1 ) {
            app.midiToBitwig( midi.status, this.noteTable[midi.data1], midi.data2 );
            this.playNotes[midi.padIndex] = -1;
        }
    }

};


/**
 * Converts cc127 to channel aftertouch.
 * @param midi
 */
Notes.prototype.onController = function( midi )
{
    if( midi.data1 === 127 && app.getMode() !== 'clips' && app.preferences.sendPressure ) {
        app.midiToBitwig( 208, midi.data2, 0 );
    }
};


Notes.prototype.getOctave    = function() { return this.getValue( 'octave' ); };
Notes.prototype.getSemitones = function() { return this.getValue( 'semitone' ); };
Notes.prototype.getScale     = function() { return Notes.scales[this.getValue( 'scale') ].notes; }


Notes.prototype.setOctave = function( octave )
{
    app.notify( 'Octave: ' + octave );
    app.preferences.octaveSetting.setRaw( octave );
    this.update();
};


Notes.prototype.setSemitones = function( semitones )
{
    app.notify( 'Root: ' + MidiData.getRootName( semitones ));
    app.preferences.semitoneSetting.setRaw( semitones );
    this.update();
};


Notes.prototype.setScale = function( scale )
{
    app.preferences.scaleSetting.setRaw( scale );
    var rootNote = MidiData.getRootName( this.sensors.semitone.value );
    var scaleName = rootNote + ' ' + Notes.scales[scale].name;
    app.notify( 'Scale: ' + scaleName );
    this.update();
};


Notes.prototype.onPreferences = function( sender, signal ) {
    this.sensors[signal].setValue( sender[signal] );
};


/**
 * Called when the app goes into alternate mode.
 * Set hold mode for pad in playing state.
 */
Notes.prototype.onEnterAlternateMode = function()
{
    var maxIndex = app.getMainMode() === 'split' ? 8 : 16;

    for( var i = 0; i < maxIndex; i++ ) {
        if( this.playNotes[i] >= 0) {
            this.holdNotes[i] = this.playNotes[i];  // set all currently playing notes to hold
            this.setHoldNote( i, 1 );
        }
    }
    this.fire( 'notesUpate' );
};


/**
 * Turns on hold state of a pad.
 * @param midi
 */
Notes.prototype.setHoldNote = function( padIndex, state )
{
    var holdPads = app.preferences.holdPads;
    if( holdPads[padIndex] !== state )
    {
        holdPads[padIndex] = state;

        app.preferences.holdSetting.set( holdPads.toString() );
        app.modeButton.onAlternateEvent();
        app.pads.updateNotePad( padIndex );
    }
};


/**
 * Sends a note-off to Bitwig for every note in this.holdNotes. Also clears the note in this.holdNotes.
 */
Notes.prototype.clearHoldNote = function( index )
{
    var note = this.holdNotes[index];
    if( note >= 0 )
    {
        app.midiToBitwig( 128, note, 0 );
        this.holdNotes[index] = -1;
    }
};






Notes.prototype.permittedModes = ['notes', 'split', 'alternate'];


Notes.scales = [
    { name: 'Chromatic',          notes: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11] },
    { name: 'Major',              notes: [0, 2, 4, 5, 7, 9, 11] },
    { name: 'Melodic Minor',      notes: [0, 2, 3, 5, 7, 9, 11] },
    { name: 'Harmonic Minor',     notes: [0, 2, 3, 5, 7, 8, 11] },
    { name: 'Gypsy Minor',        notes: [0, 2, 3, 6, 7, 8, 11] },
    { name: 'Major Pentatonic',   notes: [0, 2, 4, 7, 9] },
    { name: 'Minor Pentatonic',   notes: [0, 3, 5, 7, 10] },
    { name: 'Minor-6 Pentatonic', notes: [0, 3, 5, 7, 9] },
    { name: 'Blues',              notes: [0, 3, 5, 6, 7, 10] },
    { name: 'Japanese',           notes: [0, 1, 5, 7, 8] },
    { name: 'Hindu',              notes: [0, 2, 5, 7, 8, 11] },
    { name: 'Diminished',         notes: [0, 1, 3, 4, 6, 7, 9, 10] },
    { name: 'Whole Tone',         notes: [0, 2, 4, 6, 8, 10] },
];


