
/**
 * Pads
 * Handles the pad's functionality and LED's in all modes.
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var Pads = function()
{
    this.pendingLeds = initArray( 0, 16 );   // intermediate array that allows to check if an led color has changed
    this.currentLeds = initArray( -1, 16 );  // array for the led colors

    this.clipContent = initArray( false, 16 );
    this.clipStates  = initArray( { state: 0, queued: false }, 16 );

    app.preferences.addListener( 'scalesEnabled', this, this.updateLeds );
    app.notes.addListener( 'notesUpdate', this, this.updateLeds );

    var self = this;
    var makeContentCallback = function( slot )
    {
        return function( scene, value )
        {
            var index = self.rotationMatrix[ scene * 4 + slot ];
            self.clipContent[index] = value;
            if( app.getMode() === 'clips' || app.getMode() === 'split' ) {
                self.updateClipPad( index );
            }
        }
    };
    var makeStateCallback = function( slot )
    {
        return function( scene, state, queued )
        {
            var index = self.rotationMatrix[ scene * 4 + slot ];
            self.clipStates[index] = { state: state, queued: queued };
            if( app.getMode() === 'clips' || app.getMode() === 'split' ) {
                self.updateClipPad( index );
            }
        }
    };

    for( var i = 0; i < 4; i++ )
    {
        var channel = app.clipLauncher.getChannel( i );
        var slots = channel.getClipLauncherSlots();

        slots.addHasContentObserver( makeContentCallback( i ));
        slots.addPlaybackStateObserver( makeStateCallback( i ));
    }
};
Pads.prototype = new SensorGroup();
Pads.prototype.constructor = Pads;


Pads.prototype.onNoteOn = function( midi )
{
    if( midi.mainMode === 'clips' || midi.mainMode === 'split' )
    {
        var padIndex = Math.abs( 15 - midi.padIndex );
        var maxIndex = midi.mode === 'split' ? 8 : 16;

        if( padIndex >= 0 && padIndex < maxIndex ) {
            var index    = this.rotationMatrix[ padIndex ];
            var row      = index % 4;
            var col      = (index - row) / 4;
            var state    = this.clipStates[midi.padIndex];
            var slots    = app.clipLauncher.getTrack( row ).getClipLauncherSlots();

            if( midi.mode !== 'alternate' ) {       // launch clip
                state.state == 1 ? slots.stop() : slots.launch( col );
            }
            else {                                  // record clip
                slots.deleteClip( col );
                slots.record( col );
                app.modeButton.onAlternateEvent();
            }
        }
    }
};


Pads.prototype.onNoteOff = function( midi )
{
    if( midi.padMode === 'notes' || midi.padMode === 'clips' || midi.mode === 'alternate' )
    {
        var index = Math.abs( midi.padIndex );
        if( index < 16 ) {
            this.sendLed( index, true );
        }
    }
};


/**
 * Builds an array of color information for all pads and sends it to the pads.
 */
Pads.prototype.updateLeds = function()
{
    var mode = app.getMainMode();

    if( mode === 'notes' || mode === 'split' )
    {
        var maxIndex = mode === 'split' ? 8 : 16;
        this.pendingLeds = initArray( this.colors.dark, 16 );
        this.sendLeds( 0, maxIndex );

        for( var i = 0; i < maxIndex; i++ ) {
            this.updateNotePad( i );
        }
        this.sendLeds( 0, 16 );
    }

    if( mode === 'clips' || mode === 'split' ) {
        for( var i = 0; i < 16; i++ ) {
            this.updateClipPad( i );
        }
    }
};


if (!Array.prototype.last){
    Array.prototype.last = function(){
        return this[this.length - 1];
    };
};


/**
 * Updates a single pad for note mode.
 * @param index
 */
Pads.prototype.updateNotePad = function( index )
{
    var scale       = app.notes.getScale();
    var scaleLength = scale.length;

    if( app.preferences.scalesEnabled )
    {
        var color = index % scaleLength === 0 ? this.colors.red : this.colors.green;
        this.pendingLeds[index] = color;
    }
    else {
        var maxIndex = app.getMainMode() === 'split' ? 8 : 16;
        var octave   = Math.floor( index / scaleLength );
        var degree   = index % scaleLength;
        var pad      = scale[degree] + ( octave * 12 );
        var color    = degree === 0 ? this.colors.red : this.colors.green;

        if( pad < maxIndex && pad >= 0 ) {
            this.pendingLeds[pad] = color;
        }
    }

    if( app.preferences.holdPads[index] > 0 ) {
        this.pendingLeds[index] = this.colors.orange;
    }
    this.sendLed( index, true );
};


/**
 * Updates a single pad for clip launcher mode.
 * @param index
 */
Pads.prototype.updateClipPad = function( index )
{
    var minIndex = app.getMainMode() === 'clips' ? 0 : 8;

    if( index >= minIndex ) {
        this.pendingLeds[index] = this.colors.dark;

        if( this.clipContent[index] )            this.pendingLeds[index] = this.colors.yellow;
        if( this.clipStates[index].state === 1 ) this.pendingLeds[index] = this.colors.green;
        if( this.clipStates[index].state === 2 ) this.pendingLeds[index] = this.colors.red;
        if( this.clipStates[index].queued )      this.pendingLeds[index] = this.colors.orange;

        this.sendLed( index, true );
    }
};


/**
 * Sends the color codes for all LEDs between min and max.
 * If an LED already has the color sending is suppressed.
 * @param min first pad (0..16)
 * @param max last pad (0..16)
 * @param forceSend If true the code is sent anyway
 */
Pads.prototype.sendLeds = function( min, max, forceSend )
{
    for( i = min; i < max; i++ ) {
        this.sendLed( i, forceSend );
    }
};


/**
 * Sends the color code for the LED with the given index.
 * If the LED already has the color sending is suppressed.
 * @param index
 * @param forceSend If true the code is sent anyway
 */
Pads.prototype.sendLed = function( index, forceSend )
{
    if( index > 15 ) throw 'Pads.sendLed: index out of range ('+index+')';

    if( this.pendingLeds[index] != this.currentLeds[index] || forceSend )
    {
        this.currentLeds[index] = this.pendingLeds[index];
        var color  = this.currentLeds[index];
        var first  = 0;
        var second = 0;

        switch( color ) {
            case this.colors.red    : first = 0;    second = 127; break;
            case this.colors.green  : first = 127;  second = 0;   break;
            case this.colors.yellow : first = 127;  second = 5;   break;
            case this.colors.orange : first = 127;  second = 127; break;
        }
        app.midiToQuneo( 144, [ index * 2 ], first, 'Pads.sendLed' );
        app.midiToQuneo( 144, [ index * 2 + 1 ], second, 'Pads.sendLed'  );
    }
};





/**
 * Quneo pads are layed out this way:
 *
 * 12, 13, 14, 15
 *  8, 9,  10, 11
 *  4, 5,   6,  7
 *  0, 1,   2,  3
 *
 *  Using clip launcher indexing displays with this display tracks as columns. This in uneasy to work with,
 *  since in Bitwig's launcher view tracks are rows. The rotationMatrix rotates the indices
 *  90 degree clockwise.
 */
Pads.prototype.rotationMatrix = [
    12,  8, 4, 0,
    13,  9, 5, 1,
    14, 10, 6, 2,
    15, 11, 7, 3
];


Pads.prototype.colors = {
    dark   : 0,
    red    : 1,
    green  : 2,
    yellow : 3,
    orange : 4
};


