
/**
 * Preferences
 * Handles the Bitwig Controller Preferences.
 * Hidden preferences provide state persistence between sessions.
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var Preferences = function()
{
    this.mode              = 'notes';
    this.splitMode         = false;
    this.topSliderTarget   = 0;
    this.fixedSliderTarget = 0;
    this.otherSliderTarget = 0;
    this.sendPressure      = true;
    this.monitorEnabled    = false;
    this.messagesEnabled   = true;
    this.octave            = 0;
    this.semitone          = 0;
    this.scale             = 0;
    this.holdPads          = initArray( 0, 16 );
    this.scalesEnabled     = true;

    var prefs = host.getPreferences();

    this.scalesEnabledSetting = prefs.getEnumSetting(
        'Scales', 'Pads', ['Off', 'On'], this.scalesEnabled ? 'On' : 'Off'
    );
    this.sendPressureSetting = prefs.getEnumSetting(
        'Send Pressure', 'Pads', ['Off', 'On'], this.sendPressure ? 'On' : 'Off'
    );
    this.monitorSetting = prefs.getEnumSetting(
        'Midi Monitor', 'Console', ['Off', 'On'], this.monitorEnabled ? 'On' : 'Off'
    );
    this.messagesSetting = prefs.getEnumSetting(
        'Messages', 'Console', ['Off', 'On'], this.messagesEnabled ? 'On' : 'Off'
    );

    // These hidden prefs are used to persist values during sessions.
    this.modeSetting = prefs.getEnumSetting(
        'Mode', '', ['Notes', 'Clips'], this.mode.capitalize()
    );
    this.splitModeSetting = prefs.getEnumSetting(
        'Split Mode', '', ['Off', 'On'], this.splitMode ? 'On' : 'Off'
    );
    this.topSliderTargetSetting = prefs.getNumberSetting(
        'Top Slider Target: ', '', 0, 5, 1, '', this.topSliderTarget
    );
    this.fixedSliderTargetSetting = prefs.getNumberSetting(
        'Fixed Slider Target: ', '', 0, 3, 1, '', this.fixedSliderTarget
    );
    this.otherSliderTargetSetting = prefs.getNumberSetting(
        'Other Slider Target: ', '', 0, 17, 1, '', this.otherSliderTarget
    );
    this.holdSetting = prefs.getStringSetting(
        'Hold Pads', '', 48, this.holdPads.toString()
    );
    this.octaveSetting = prefs.getNumberSetting(
        'Octave', '', -4, 4, 1, '', this.octave
    );
    this.semitoneSetting = prefs.getNumberSetting(
        'Semitone', '', -12, 12, 1, '', this.semitone
    );
    this.scaleSetting = prefs.getNumberSetting(
        'Scale', '', 0, app.getScales().length-1, 1, '', this.scale
    );

    this.modeSetting.hide();
    this.splitModeSetting.hide();
    this.topSliderTargetSetting.hide();
    this.fixedSliderTargetSetting.hide();
    this.otherSliderTargetSetting.hide();
    this.octaveSetting.hide();
    this.semitoneSetting.hide();
    this.scaleSetting.hide();
    this.scalesEnabledSetting.hide();
    this.holdSetting.hide();

    Observer.makeObserver( this );

    var self = this;

    this.scalesEnabledSetting.addValueObserver( function( value ) {
        value = ( value === 'On' );
        if( self.scalesEnabled !== value ) {
            self.scalesEnabled = value;
            self.fire( 'scalesEnabled' );
        }
    });

    this.sendPressureSetting.addValueObserver( function( value ) {
        value = ( value === 'On' );
        if( self.sendPressure !== value ) {
            self.sendPressure = value;
            self.fire( 'sendPressure' );
        }
    });

    this.modeSetting.addValueObserver( function( value ) {
        value = value.toLowerCase();
        if( self.mode !== value ) {
            self.mode = value;
            self.fire( 'mode' );
        }
    });

    this.splitModeSetting.addValueObserver( function( value ) {
        value = !!(value === 'On');
        if( self.splitMode !== value ) {
            self.splitMode = value;
            self.fire( 'splitMode' );
        }
    });

    this.topSliderTargetSetting.addRawValueObserver( function( value ) {
        if( self.topSliderTarget !== value ) {
            self.topSliderTarget = value;
            self.fire( 'topSliderTarget' );
        }
    });

    this.fixedSliderTargetSetting.addRawValueObserver( function( value ) {
        if( self.fixedSliderTarget !== value ) {
            self.fixedSliderTarget = value;
            self.fire( 'fixedSliderTarget' );
        }
    });

    this.otherSliderTargetSetting.addRawValueObserver( function( value ) {
        if( self.otherSliderTarget !== value ) {
            self.otherSliderTarget = value;
            self.fire( 'otherSliderTarget' );
        }
    });

    this.messagesSetting.addValueObserver( function( value ) {
        value = ( value === 'On' );
        if( self.messagesEnabled !== value ) {
            self.messagesEnabled = value;
            self.fire( 'messagesEnabled' );
        }
    });

    this.monitorSetting.addValueObserver( function( value ) {
        value = ( value === 'On' );
        if( self.monitorEnabled !== value ) {
            self.monitorEnabled = value;
            self.fire( 'monitorEnabled' );
        }
    });

    this.holdSetting.addValueObserver( function( value ) {
        if( self.holdPads.toString() !== value ) {
            self.holdPads = value.split( ',' );
            self.fire( 'holdPads' );
        }
    });

    this.octaveSetting.addRawValueObserver( function( value ) {
        if( self.octave !== value ) {
            self.octave = value;
            self.fire( 'octave' );
        }
    });

    this.semitoneSetting.addRawValueObserver( function( value ) {
        if( self.semitone !== value ) {
            self.semitone = value;
            self.fire( 'semitone' );
        }
    });

    this.scaleSetting.addRawValueObserver( function( value ) {
        if( self.scale !== value ) {
            self.scale = value;
            self.fire( 'scale' );
        }
    });
};





