/**
 * Router
 * Routes the slider to their targets.
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var Router = function()
{
    var self     = this;
    this.sliders = new Sliders();

    this.sensors.topSelector   = new RangedSensorPair( 0, 0, 2, 'topTargetUp', 'topTargetDown' );
    this.sensors.fixedSelector = new RangedSensorPair( 0, 0, 1, 'subTargetUp', 'subTargetDown' );
    this.sensors.paramSelector = new AutomatedSensorPair( 'subTargetUp', 'subTargetDown' );
    this.sensors.otherSelector = new RangedSensorPair( 0, 0, 17, 'subTargetUp', 'subTargetDown' );

    this.sensors.topSelector.addListener( 'sensorOn', this, function( sensor ) {
        app.preferences.topSliderTargetSetting.setRaw( sensor.getValue() );
        this.notifyApp();
        this.updateLeds();
    });
    this.sensors.fixedSelector.addListener( 'sensorOn', this, function( sensor ) {
        app.preferences.fixedSliderTargetSetting.setRaw( sensor.getValue() );
        this.notifyApp();
        this.updateLeds();
    });
    this.sensors.paramSelector.addListener( 'sensorInc', this, function() {
        app.device.nextParameterPage();
        for( var i = 0; i < 8; i++ ) {
            app.device.getParameter(i).setIndication( true );
        }
    });
    this.sensors.paramSelector.addListener( 'sensorDec', this, function() {
        app.device.previousParameterPage();
        for( var i = 0; i < 8; i++ ) {
            app.device.getParameter(i).setIndication( true );
        }
    });
    this.sensors.otherSelector.addListener( 'sensorOn', this, function( sensor ) {
        app.preferences.otherSliderTargetSetting.setRaw( sensor.getValue() );
        this.notifyApp();
        this.updateLeds();
    });


    // Load state from the preferences
    app.preferences.addListener( 'topSliderTarget', this, function() {
        if( this.sensors.topSelector.getValue() !== app.preferences.topSliderTarget ) {
            this.sensors.topSelector.setValue( app.preferences.topSliderTarget );
        }
    });
    app.preferences.addListener( 'fixedSliderTarget', this, function() {
        this.sensors.fixedSelector.setValue( app.preferences.fixedSliderTarget );
    });
    app.preferences.addListener( 'otherSliderTarget', this, function() {
        this.sensors.otherSelector.setValue( app.preferences.otherSliderTarget );
    });


    // Enable/disable buttons
    app.device.addNextParameterPageEnabledObserver( function( enabled ) {
        self.sensors.paramSelector.enableIncrease( enabled, self.getTarget().key === 'parameter' );
    });
    app.device.addPreviousParameterPageEnabledObserver( function( enabled ) {
        self.sensors.paramSelector.enableDecrease( enabled, self.getTarget().key === 'parameter' );
    });

    this.targets = [
        {
            selector: self.sensors.fixedSelector,
            subTargets: [{
                key    : 'common',
                label  : 'Fixed Sections Common Controls',
                values : initArray( 0, 8 ),
                context: app.device,
                device : app.device.getCommonParameter
            }, {
                key    : 'envelope',
                label  : 'Fixed Sections Envelope Controls',
                values : initArray( 0, 8 ),
                context: app.device,
                device : app.device.getEnvelopeParameter
            }]
        }, {
            selector: self.sensors.paramSelector,
            subTargets: [{
                key    : 'parameter',
                label  : 'Parameter Pages',
                values : initArray( 0, 8 ),
                context: app.device,
                device : app.device.getParameter
            }],
        }, {
            selector: self.sensors.otherSelector,
            subTargets: [{
                key    : 'macro',
                label  : 'Macros',
                values : initArray( 0, 8 ),
                context: app.device,
                device : function( index ) {
                    return app.device.getMacro( index ).getAmount();
                }
            }, {
                key    : 'send',
                label  : 'Sends',
                values : initArray( 0, 8 ),
                context: app.cursorTrack,
                device : app.cursorTrack.getSend
            }]
        }
    ];
    for( var i = 0; i < 16; i++ ) {     // 16 banks of controller targets
        this.targets[2].subTargets[i+2] = {
            key    : 'ctrl',
            label  : 'CC '+(i*8+1)+'-'+(i*8+8),
            values : initArray( 0, 8 ),
            context: self,
            device : function( index ) {
                return {
                    set: function( value ) { self.sendController( index, value ); },
                    reset: function()      { self.sendController( index, 0 ); }
                };
            }
        };
    }

    // Callbacks on target changes
    var makeCallback = function( index, source, target )
    {
        source.addValueObserver( 128, function( value ) {
            target.values[index] = value;
            if( self.getTarget() === target ) {
                self.sliders.setValue( index, value, true );
            }
        });
    };
    for( var i = 0; i < 8; i++ ) {
        makeCallback( i, app.device.getCommonParameter(i),   this.targets[0].subTargets[0] );
        makeCallback( i, app.device.getEnvelopeParameter(i), this.targets[0].subTargets[1] );
        makeCallback( i, app.device.getParameter(i),         this.targets[1].subTargets[0] );
        makeCallback( i, app.device.getMacro(i).getAmount(), this.targets[2].subTargets[0] );
        makeCallback( i, app.cursorTrack.getSend(i),         this.targets[2].subTargets[1] );

        // callbacks for controller target
        this.sliders.addListener( i, 'sensorMove', this, function( slider ) {
            var target = self.getTarget();
            if( target.key === 'ctrl' ) {
                target.values[slider.getIndex()] = slider.getValue();
            }
        });
    }

    // UserControls
    this.userControls = host.createUserControls( 128 );
    for( var i = 1; i < 128; i++ ) {
        this.userControls.getControl( i ).setLabel( 'CC-' + i );
    }
};
Router.prototype = new SensorGroup();
Router.prototype.constructor = Router;


Router.prototype.onNoteOn = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.topTargetDown.note  : this.sensors.topSelector.decrease(); break;
        case app.layout.topTargetUp.note    : this.sensors.topSelector.increase(); break;
        case app.layout.subTargetUp.note    : this.getSubSelector().increase(); break;
        case app.layout.subTargetDown.note  : this.getSubSelector().decrease(); break;
    }
    this.sliders.onNoteOn( midi );
};


Router.prototype.onNoteOff = function( midi )
{
    switch( midi.data1 ) {
        case app.layout.topTargetUp.note    :
        case app.layout.topTargetDown.note  : this.sensors.topSelector.onRelease(); break;
        case app.layout.subTargetUp.note    :
        case app.layout.subTargetDown.note  : this.getSubSelector().onRelease(); break;
    }
    this.sliders.onNoteOff( midi );
};


Router.prototype.onController = function( midi ) {
    this.sliders.onController( midi );
};


/**
 * Update all button and slider leds.
 */
Router.prototype.updateLeds = function()
{
    this.sensors.topSelector.sendLed();
    this.getSubSelector().sendLed();
    this.sliders.setValues( this.getTarget().values );
};


Router.prototype.getTarget = function()
{
    var topIndex  = this.sensors.topSelector.getValue();
    var topTarget = this.targets[topIndex];
    var subIndex  = topIndex === 1 ? 0 : topTarget.selector.getValue();

    return topTarget.subTargets[subIndex];
};


/**
 * Returns the current sub target selector for the current top target.
 * @returns {*}
 */
Router.prototype.getSubSelector = function() {
    return this.targets[this.sensors.topSelector.getValue() ].selector;
};


/**
 * Sets a target value.
 * @param index Index of active sub target
 * @param value The new value for the target value
 */
Router.prototype.setValue = function( index, value )
{
    var target = this.getTarget();
    var device = target.device.call( target.context, index );
    device.set( value, 128 );
};


/**
 * Resets a target value to it's default.
 * @param index Index of active sub target
 */
Router.prototype.resetValue = function( index )
{
    var target = this.getTarget();
    var device = target.device.call( target.context, index );
    device.reset();
};


/**
 * Sends a Midi controller to Bitwig.
 * @param sliderIndex Index of originating slider
 * @param value Value to send
 */
Router.prototype.sendController = function( sliderIndex, value )
{
    var base          = (this.sensors.otherSelector.getValue() - 2) * 8 + 1;
    var controllerNum = base + sliderIndex;
    host.showPopupNotification( 'CC-'+controllerNum+': '+value );

    this.userControls.getControl( controllerNum ).set( value, 128 );
    app.midiToBitwig( 176, controllerNum, value );
};


/**
 * Calls app.notify with the label of current target.
 */
Router.prototype.notifyApp = function() {
    app.notify( 'Sliders: ' + this.getTarget().label );
};

Router.prototype.permittedModes    = [ 'notes', 'clips', 'split' ];











