/**
 * Sensor
 * Base class for all sensors.
 *
 * Sensor classes wrap access to Quneo sensors.
 * They abstract midi events to the following set of events, to which listeners can register:
 * - sensorOn: a sensor was pressed
 * - sensorOff: a sensor was released
 * - sensorMove: a sensor was moved.
 * - doublePress: a sensor was pressed twice with 500 ms.
 * Also sensor class maintain a value, which they convert to a Quneo led status and send it to Quneo.
 *
 * @param layoutKey {string} A string key to retrieve the layout data from app.layout.
 *
 * @author 3ears
 * @mail tom-buettner@t-online.de
 */
var Sensor = function( layoutKey, value )
{
    Observer.makeObserver( this );

    this.layoutKey = layoutKey;
    this.value     = value || 0;
};


/**
 * Call this method on midi events that signal a sensor press. Fires the sensorOn-event.
 */
Sensor.prototype.onPress = function()
{
    this.waitForDoublePress();
    this.fire( 'sensorOn' );
};

/**
 * Call this method on midi events that signal a sensor release. Fires the sensorOff-event.
 */
Sensor.prototype.onRelease = function()
{
    this.sendLed();
    this.fire( 'sensorOff' );
};

/**
 * Call this method on midi events that signal a sensor move. Fires the sensorMove-event.
 * @param value The new value for the sensor
 */
Sensor.prototype.onMove = function( value )
{
    this.value = value;
    this.fire( 'sensorMove' );
};


/**
 * Sets a new value for the sensor.
 * @param value The new value
 * @param update If true, the leds for the sensor are updated immediately.
 */
Sensor.prototype.setValue = function( value, update )
{
    this.value = value;
    if( update ) this.sendLed();
};

/**
 * Returns the sensor value.
 * @returns {*|number}
 */
Sensor.prototype.getValue = function() { return this.value; }


/**
 * Converts the sensor value into Quneo led code.
 * @param value The value to convert. If undefined this.value is used.
 * @param layoutKey The key to retrieve the led layout. If undefined this.layoutKey is used.
 */
Sensor.prototype.sendLed = function( value, layoutKey )
{
    var layout = this.getLayout( layoutKey );
    value = (typeof value === 'undefined') ? this.value : value;

    app.midiToQuneo( layout.ledStatus, layout.ledNum, value ? 127 : 0, layoutKey || this.layoutKey );
};

/**
 * Retrieves the led layout from app.layout.
 * @param layoutKey The key for app.layout. If undefined this.layoutKey is used.
 * @returns An entry from app.layout
 */
Sensor.prototype.getLayout = function( layoutKey ) {
    return app.layout[this.layoutKey || layoutKey];
};


/**
 * Implements double press functionality for sensors.
 * If two sensorOn events occur with 500 ms a doublePress event is fired.
 * After a doublePress event the sensor will not fire further events for 1000 ms.
 */
Sensor.prototype.waitForDoublePress = function()
{
    if( this.firstPress )
    {
        this.firstPress = false;
        this.fire( 'doublePress' );
        this.canFire = false;
        host.scheduleTask( utils.bind( function() { // block sending for while to avoid setting the value after reset
            this.canFire = true;
        }, this ), null, 1000 );
    }
    else {
        this.firstPress = true;
        host.scheduleTask( utils.bind( function() {
            this.firstPress = false;
        }, this ), null, 500 );
    }
};


/**
 * AlwaysOnSensor
 * A sensor class that always displays on-state.
 * @param layoutKey
 * @param value
 * @constructor
 */
var AlwaysOnSensor = function( layoutKey, value )
{
    Sensor.prototype.constructor.call( this, layoutKey, value );
};
AlwaysOnSensor.prototype = new Sensor();
AlwaysOnSensor.prototype.constructor = AlwaysOnSensor;


AlwaysOnSensor.prototype.sendLed = function( value ) {
    Sensor.prototype.sendLed.call( this, 127 );
};


/** --------------------------------------------------------------------------
 * SensorPair
 * Abstract base for classes the handle the Quneo button pairs.
 * @param incLayout layoutKey for the right (or top button)
 * @param decLayout layoutKey for left (or bottom button)
 * @constructor
 */
var SensorPair = function( incLayout, decLayout )
{
    this.incLayout = incLayout;
    this.decLayout = decLayout;
    this.ledValues = { inc: 0, dec: 0 };
    Observer.makeObserver( this );
};
SensorPair.prototype = new Sensor();
SensorPair.prototype.constructor = SensorPair;


SensorPair.prototype.increase = function() {}
SensorPair.prototype.decrease = function() {}
SensorPair.prototype.setValue = function( value ) {};


SensorPair.prototype.sendLed = function( value )
{
    var useOwn = typeof value === 'undefined';
    Sensor.prototype.sendLed.call( this, useOwn ? this.ledValues.inc : value, this.incLayout );
    Sensor.prototype.sendLed.call( this, useOwn ? this.ledValues.dec : value, this.decLayout );
};


/**
 * RangedSensorPair
 * A sensor pair that moves a value within a minimum and a maximum
 * @param value The initial value
 * @param min Minimum value
 * @param max Maximum value
 * @param incLayout
 * @param decLayout
 * @param bipolar If true the leds will work bipolar ( left is on when value < 0, right is on when value > 0
 * @constructor
 */
var RangedSensorPair = function( value, min, max, incLayout, decLayout, bipolar )
{
    SensorPair.prototype.constructor.call( this, incLayout, decLayout );

    this.min     = min;
    this.max     = max;
    this.bipolar = bipolar || false;
    this.setValue( value || 0 );
};
RangedSensorPair.prototype = new SensorPair();
RangedSensorPair.prototype.constructor = RangedSensorPair;


RangedSensorPair.prototype.increase = function()
{
    this.setValue( this.value + 1 );
    this.fire( 'sensorInc' );
};


RangedSensorPair.prototype.decrease = function()
{
    this.setValue( this.value - 1 );
    this.fire( 'sensorDec' );
};


RangedSensorPair.prototype.setValue = function( value )
{
    value = Math.min( this.max, Math.max( this.min, value ));

    if( this.bipolar ) {
        var step = (127 / this.max - this.min) * Math.abs( value );
        this.ledValues = {
            dec: value < 0 ? step : 0,
            inc: value > 0 ? step : 0
        };
    }
    else {
        this.ledValues = {
            dec: value === this.min ? 0 : 127,
            inc: value === this.max ? 0 : 127
        };
    }
    if( value !== this.value ) {
        this.value = value;
        this.fire( 'sensorOn' );
    }
};



/**
 * CircleSensorPair
 * A sensor pair that works like an endless knob.
 * @param value
 * @param length
 * @param incLayout
 * @param decLayout
 * @constructor
 */

var CircleSensorPair = function( value, length, incLayout, decLayout )
{
    SensorPair.prototype.constructor.call( this, incLayout, decLayout );
    this.length = length;
    this.setValue( value );
};
CircleSensorPair.prototype = new SensorPair();
CircleSensorPair.prototype.constructor = CircleSensorPair;


CircleSensorPair.prototype.increase = function()
{
    var value = (this.value + 1) % this.length;
    this.setValue( value );
};


CircleSensorPair.prototype.decrease = function()
{
    var value = this.value;
    if( value === 0 ) {
        value = this.length - 1;
    } else {
        value = (value - 1) % this.length;
    }
    this.setValue( value );
};


CircleSensorPair.prototype.setValue = function( value )
{
    this.ledValues = {
        dec: value === 0 ? 0 : 127,
        inc: value === 0 ? 0 : 127
    };
    if( value !== this.value ) {
        this.value = value;
        this.fire( 'sensorOn' );
    }
};


//--------------------------------------------------------------
// AutomatedSensorPair
//--------------------------------------------------------------
/**
 * AutomatedSensorPair
 * The status for the sensor pair is controlled from outside by setting the incEnabled and decEnabled flags.
 * It sends sensorInc and sensorDec signals on press.
 * @param incLayout
 * @param decLayout
 * @constructor
 */

var AutomatedSensorPair = function( incLayout, decLayout )
{
    SensorPair.prototype.constructor.call( this, incLayout, decLayout );
    this.incEnabled = false;
    this.decEnabled = false;
};
AutomatedSensorPair.prototype = new SensorPair();
AutomatedSensorPair.prototype.constructor = AutomatedSensorPair;


AutomatedSensorPair.prototype.sendLed = function()
{
    Sensor.prototype.sendLed.call( this, this.incEnabled ? 127 : 0, this.incLayout );
    Sensor.prototype.sendLed.call( this, this.decEnabled ? 127 : 0, this.decLayout );
};


AutomatedSensorPair.prototype.increase = function() {
    if( this.incEnabled ) this.fire( 'sensorInc' );
};
AutomatedSensorPair.prototype.decrease = function() {
    if( this.decEnabled ) this.fire( 'sensorDec' );
};

AutomatedSensorPair.prototype.enableIncrease = function( enable, update ) {
    this.incEnabled = enable;
    if( update ) this.sendLed();
};
AutomatedSensorPair.prototype.enableDecrease = function( enable, update ) {
    this.decEnabled = enable;
    if( update ) this.sendLed();
};


/**
 * RhombusButton
 * Handles the rhombus button.
 * The rhombus button has two leds and thus can display multiple colors.
 * @param layoutKey
 * @param value
 * @param colorMap
 * @constructor
 */
var RhombusButton = function( layoutKey, value, colorMap )
{
    Sensor.prototype.constructor.call( this, layoutKey, value );
    this.colorMap = colorMap;
};
RhombusButton.prototype = new Sensor();
RhombusButton.prototype.constructor = RhombusButton;


RhombusButton.prototype.sendLed = function()
{
    var layout  = this.getLayout();
    var ledBase = layout.ledNum;
    var colors  = this.colorMap[ this.value ];

    app.midiToQuneo( layout.ledStatus, ledBase, colors[0] );
    app.midiToQuneo( layout.ledStatus, ledBase+1, colors[1] );
};



/** --------------------------------------------------------------------------
 * Slider
 * Manages a slider or a rotary
 *
 */
var Slider = function( layoutKey, index )
{
    this.index = index;
    Sensor.prototype.constructor.call( this, layoutKey );
};
Slider.prototype = new Sensor();
Slider.prototype.constructor = Slider;

Slider.prototype.getIndex = function() { return this.index; }

Slider.prototype.sendLed = function()
{
    var layout = this.getLayout( this.layoutKey );
    app.midiToQuneo( layout.ledStatus, layout.ledNum, this.value, this.layoutKey );
};


/**
 * Rotary
 * A rotary sensor that increases it's value when moved clockwise
 * and decreases it's value when move counterclockwise.
 * @param layoutKey {string} A key to retrieve the layout data from app.layout.
 * @constructor
 */
var Rotary = function( min, max, layoutKey, value )
{
    Sensor.prototype.constructor.call( this, layoutKey, value );
    this.min       = min;
    this.max       = max;
    this.fullValue = this.value;
    this.previous  = this.value;
};
Rotary.prototype = new Sensor();
Rotary.prototype.constructor = Sensor;


Rotary.prototype.onMove = function( value )
{
    this.previous = this.previous || 0;
    var delta     = value - this.previous;

    if( this.previous !== 0 ) {
        if( delta <= -100 ) delta = 0;
        else if( delta >= 100 ) delta = 0;
    }

    value = this.previous + delta;

    if( this.min !== null ) value = Math.max( this.min, value );
    if( this.max !== null ) value = Math.min( this.max, value );
    this.previous = value;

    if( value !== this.value ) {
        this.fullValue = value
        this.value     = value % 128;
        this.fire( 'sensorMove' );
    }
};


Rotary.prototype.sendLed = function()
{
    var layout = this.getLayout( this.layoutKey );
    app.midiToQuneo( layout.ledStatus, layout.ledNum, this.value, this.layoutKey );
};


Rotary.prototype.getValue = function() {
    return this.fullValue;
};




/** -----------------------------------------------------------------------------
 * SensorGroup
 * Container class for sensors
 */
var SensorGroup = function()
{
    this.sensors = {};
};


SensorGroup.prototype.updateLeds = function()
{
    for( var key in this.sensors ) {
        this.sensors[key].sendLed();
    }
};


SensorGroup.prototype.getValue = function( sensorName ) {
    return this.sensors[sensorName].getValue();
};


SensorGroup.prototype.setValue = function( sensorName, value, update ) {
    this.sensors[sensorName].setValue( value, update );
};


SensorGroup.prototype.addListener = function( sensorName, signal, context, callback ) {
    this.sensors[sensorName].addListener( signal, context, callback );
};


SensorGroup.prototype.onMidi = function( midi )
{
    if( this.isActive() ) {
        if( midi.isNoteOn() )          this.onNoteOn( midi );
        else if( midi.isNoteOff() )    this.onNoteOff( midi );
        else if( midi.isController() ) this.onController( midi );
    }
};


/**
 * Checks if the SensorGroup has permission for the current mode.
 */
SensorGroup.prototype.isActive = function()
{
    var result = false;
    var length = this.permittedModes.length;
    if( length === 0 ) result = true;
    if( app.isAlternateMode() && this.permittedModes.indexOf( 'alternate' ) > -1 ) result = true;
    if( result === false ) {
        var mode = app.getMode();
        for( var i = 0; i < length; i++ ) {
            if( mode === this.permittedModes[i] ) result = true;
        }
    }
    return result;
};

SensorGroup.prototype.permittedModes = [];

SensorGroup.prototype.onNoteOn     = function( midi ) {}
SensorGroup.prototype.onNoteOff    = function( midi ) {}
SensorGroup.prototype.onController = function( midi ) {}
